import pygraphviz as pgv
from pykrusch.parsePymc.randomVariable import specify_rv_nodes, add_rv_nodes
from pykrusch.parsePymc.deterministicVariable import specify_f_nodes, add_f_nodes
from pymc import Model


def create_graph(model: Model, plot_posterior, posterior_data):
    """
    This function serves as a 'main' of sorts: it is resposible for calling
    the other functions needed to get all the nodes in the graph
    and then assemble them into graph format.
    """

    # Start by finding the parameters for each of the
    # free, observed, and deterministic nodes
    rv_nodes = specify_rv_nodes(model.free_RVs, plot_posterior, posterior_data)
    obs_nodes = specify_rv_nodes(
        model.observed_RVs, plot_posterior=False, posterior_data=None
    )
    f_nodes = specify_f_nodes(model.deterministics)

    # Instantiate an empty directed graph
    graph = pgv.AGraph(directed=True, strict=False)

    # Add the nodes to the graph
    add_rv_nodes(rv_nodes, graph)
    add_rv_nodes(obs_nodes, graph)
    add_f_nodes(f_nodes, graph)

    return graph



from __future__ import annotations
from PIL import Image
from typing import TYPE_CHECKING
from pygraphviz import AGraph
from pykrusch.config import *

if TYPE_CHECKING:
    from pykrusch.figureControl import MathImage
    from pykrusch.parsePymc.randomVariable import RandomVar
    from pykrusch.parsePymc.deterministicVariable import DeterministicVar
    from pykrusch.dist import Dist
    from pykrusch.param import NamedParameter, NamedData


def make_data_html(param: NamedData, graph):
    label_start = f"""<
    <TABLE BORDER="20" CELLBORDER="2" CELLSPACING="0" VALIGN="BOTTOM">
    <TR>
    """
    label_end = f"""</TR></TABLE>>"""

    label = """"""

    label += label_start

    image = Image.open(param.mi.filepath)
    tw = image.width * DETERMINISTIC_TEXT_SCALE
    th = image.height * DETERMINISTIC_TEXT_SCALE
    label += f"""<TD FIXEDSIZE="TRUE" WIDTH="{tw}" HEIGHT="{th}" PORT="{param.mi.portnum}"><IMG SCALE="BOTH" SRC="{param.mi.filepath}"></IMG></TD>"""

    label += label_end

    return label


def make_f_html(img_list: list[MathImage], graph: AGraph, f_node: DeterministicVar):
    label_start = f"""<
    <TABLE BORDER="20" CELLBORDER="2" CELLSPACING="0" VALIGN="BOTTOM">
    <TR>
    """
    label_end = f"""</TR></TABLE>>"""

    label = """"""

    label += label_start

    for img in img_list:
        image = Image.open(img.filepath)
        tw = image.width * DETERMINISTIC_TEXT_SCALE
        th = image.height * DETERMINISTIC_TEXT_SCALE
        label += f"""<TD FIXEDSIZE="TRUE" WIDTH="{tw}" HEIGHT="{th}" PORT="{img.portnum}"><IMG SCALE="BOTH" SRC="{img.filepath}"></IMG></TD>"""
        for edge_from in img.edges_from:
            if edge_from and graph.has_edge(edge_from, f_node.name, key=img.op_id):
                e = graph.get_edge(edge_from, f_node.name, key=img.op_id)
                e.attr["headport"] = str(img.portnum) + ":n"
                e.attr["tailport"] = "s"

    label += label_end

    return label


def make_rv_html(rv: RandomVar, graph: AGraph):
    dist: Dist = rv.dist
    tw = DIST_IMAGE_WIDTH * IMAGE_SCALE_FACTOR
    th = DIST_IMAGE_HEIGHT * IMAGE_SCALE_FACTOR

    param_slots = max(dist.num_params, 1)
    slot_width = tw // param_slots

    label_start = f"""<<TABLE BORDER="1" CELLBORDER="0" CELLSPACING="0">
                    <TR>"""

    for param in dist.params:
        param_symbol, font_size = param.give_symbol()

        label_start += f"""
        <TD WIDTH="{slot_width}" PORT="{str(param.slot)}">
            <TABLE CELLPADDING="0" BORDER="0" CELLSPACING="0">
                <TR>
                    <TD><FONT POINT-SIZE="{font_size}">{param_symbol}</FONT></TD>
                </TR>
                <TR>
                    <TD><FONT POINT-SIZE="{DIST_FONT_SIZE_SMALL}">{param.meaning}</FONT></TD>
                </TR>
            </TABLE>
        </TD>"""

    label_start += """</TR>"""

    distimage = dist.distimage

    if distimage:
        label_start += f"""
                <TR>
                    <TD COLSPAN="{param_slots}" WIDTH="{tw}" HEIGHT="{th}"><IMG SCALE="TRUE" SRC="{distimage}"></IMG></TD>
                </TR>"""

    image = Image.open(dist.dist_name_latex)
    tw = image.width * DETERMINISTIC_TEXT_SCALE
    th = image.height * DETERMINISTIC_TEXT_SCALE
    label_end = f"""
                <TR>
                    <TD COLSPAN="{param_slots}" ALIGN="center" HEIGHT="25"><FONT POINT-SIZE="{DIST_FONT_SIZE_SMALL}">{dist.type.upper()}</FONT></TD>
                    
                </TR>
                <TR>
                    <TD WIDTH="{tw}" HEIGHT="{th}" COLSPAN="{param_slots}" ALIGN="center"><IMG SCALE="TRUE" SRC="{dist.dist_name_latex}"></IMG></TD>    
                </TR>
                
                
            </TABLE>>"""

    label_start += label_end

    return label_start

    # <TR>
    #     <TD COLSPAN="{param_slots}" ALIGN="center"><FONT POINT-SIZE="{DIST_FONT_SIZE}">{rv.name}</FONT></TD>
    # </TR>



DATA_LETTER = "X"
LATEX_NAMES = True

PARAM_CONVERSION = {
    "mu": "μ",
    "lam": "λ",
    "sigma": "σ",
    "alpha": "α",
    "beta": "β",
    "gamma": "γ",
    "nu": "ν",
    "tau": "τ",
    "pi": "π",
    "lambda": "λ",
    "theta": "θ",
}

NODE_COLOUR_GV = "grey97"
NODE_COLOUR_MPL = "#F7F7F7"

PARAM_FONT_SIZE = "35.0"
PARAM_FONT_SIZE_NUMERICAL = "30.0"
DIST_FONT_SIZE = "40.0"
DIST_FONT_SIZE_SMALL = "20.0"

IMAGE_SCALE_FACTOR = 1
DIST_IMAGE_WIDTH = 82.5
DIST_IMAGE_HEIGHT = 61.5

DISTROGRAM_LINE_WIDTH = 5
DISCRETE_LINE_WIDTH = 2


if LATEX_NAMES:
    DETERMINISTIC_TEXT_SCALE = 1
else:
    DETERMINISTIC_TEXT_SCALE = 0.7
DETERMINISTIC_TEXT_HEIGHT = 80

TEMP_IMG_DIR = ".PYKRUSCHFIGTEMP"

DPI = 300
ENGINE = "dot"

LINSPACE_N = 500



from pykrusch.figureControl import FC, MathImageFound
from pykrusch.graphviz.latexText import (
    render_img_from_latex,
    string_to_latex,
)
from pykrusch.graphviz.pgvHTML import make_f_html
from pytensor.graph.basic import Variable
from pygraphviz import AGraph
from pykrusch.config import DATA_LETTER, NODE_COLOUR_GV
import itertools


class DeterministicVar:
    def __init__(self, node: Variable):
        self.name: str = str(node.name)
        self.expression = start_recursive_f_search(node)

        self.shape = "plaintext"
        self.color = NODE_COLOUR_GV
        self.style = "filled"

    def f_add_self_to_graph(self, graph: AGraph):
        img_list = self.self_img_equals()
        img_list += self.expression.math_img_list()
        self.expression.op_to_graph(graph, self.name)
        graph.add_node(
            self.name,
            shape=self.shape,
            color=self.color,
            style=self.style,
            label=make_f_html(img_list, graph, self),
        )

    def self_img_equals(self):
        self_sanitized_name = self.name.replace("_", "\_")
        latex_str = string_to_latex(self_sanitized_name)
        mi_self = FC.fig_path(FC, symbols=latex_str)
        if not isinstance(mi_self, MathImageFound):
            render_img_from_latex(latex_str, mi_self.filepath)
        latex_str_eq = string_to_latex(r"=")
        mi_equals = FC.fig_path(FC, symbols=latex_str_eq)
        if not isinstance(mi_equals, MathImageFound):
            render_img_from_latex(latex_str_eq, mi_equals.filepath)
        return [mi_self, mi_equals]


class MathOP:
    OP_ID = itertools.count()

    def __init__(self, apply_parents, single_arg=False):
        self.op = []
        self.arg1 = recursive_f_search(apply_parents[0])
        if single_arg:
            self.arg2 = None
        else:
            self.arg2 = recursive_f_search(apply_parents[1])
        self.pre = []
        self.post = []
        self.op_id = "math" + str(MathOP.OP_ID.__next__())
        self.print_arg2 = True

    def __repr__(self):
        if self.arg2:
            return str(f" [{self.arg1} {self.op} {self.arg2}] ")

        else:
            return str(f" [{self.op}{self.arg1}] ")

    def __str__(self):
        return self.__repr__()

    def op_to_graph(self, graph, f_node_name):
        self.arg1.op_to_graph(graph, f_node_name)
        if self.arg2:
            self.arg2.op_to_graph(graph, f_node_name)

    def math_img_list(self) -> list:
        img_list = []
        for i in self.pre:
            img_list.append(self.self_img(i))
        img_list += self.arg1.math_img_list()
        for ii in self.op:
            img_list.append(self.self_img(ii))
        if self.arg2 and self.print_arg2:
            img_list += self.arg2.math_img_list()
        for iii in self.post:
            img_list.append(self.self_img(iii))
        return img_list

    def self_img(self, symbols):
        latex_str = string_to_latex(symbols)
        mi = FC.fig_path(FC, symbols=latex_str)
        if not isinstance(mi, MathImageFound):
            render_img_from_latex(latex_str, mi.filepath)
        return mi

    def latex_as_data(self):
        raise NotImplementedError

    def latex_as_subset(self):
        raise NotImplementedError

    def latex_as_variable(self):
        raise NotImplementedError


### Classes that inherit from MathOps
# region


class Add(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents)
        self.op = ["+"]


class Subtract(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents)
        self.op = ["-"]


class Multiply(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents)


class Divide(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents)
        self.pre = ["("]
        self.op = [")", "/", "("]
        self.post = [")"]


class Power(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents)

        n = "x"

        try:
            n = self.arg2.value
            if n.is_integer():
                n = int(n)
            self.print_arg2 = False
            self.pre = ["("]
            self.post = [f")^{{{n}}}"]
            return
        except AttributeError:
            pass
        try:
            n = self.arg2.name
            self.pre = ["("]
            self.op = [")","^\wedge","("]
            self.post = [f")"]
            return
        except AttributeError:
            pass

        self.print_arg2 = False
        self.pre = ["("]
        self.post = [f")^{{{n}}}"]


class Exp(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents, single_arg=True)
        self.pre = [
            r"\mathrm{exp}",
            "(",
        ]
        self.post = [")"]


class SubSet(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents)
        self.op = []

    # Even if the variable subset is named, we don't want
    # to create a node for it in the graph, so we only
    # ever pass the recursive function along to the first
    # node.
    def op_to_graph(self, graph, f_node_name):
        self.arg1.op_to_graph(graph, f_node_name)

    # Subset is going to have to be special
    def math_img_list(self) -> list:
        return [self.self_img()]

    def self_img(self):
        name1 = self.arg1.latex_as_variable()
        name2 = self.arg2.latex_as_subset()

        if r"_" in name1 and r"{" in name1 and r"}" in name1:
            latex_str = name1.replace(r"}", "") + rf"[{name2}]" + r"}"
        else:
            latex_str = name1 + rf"_{{[{name2}]}}"

        mi = FC.fig_path(
            FC, edges_from=[self.arg1.name], symbols=latex_str, op_id=self.arg1.op_id
        )
        if not isinstance(mi, MathImageFound):
            render_img_from_latex(latex_str, mi.filepath)

        return mi


class Logarithm(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents, single_arg=True)
        self.pre = [
            "ln",
            "(",
        ]
        self.post = [")"]


class Logarithm2(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents, single_arg=True)
        self.pre = [
            "log_2",
            "(",
        ]
        self.post = [")"]


class Logarithm10(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents, single_arg=True)
        self.pre = [
            r"log_10",
            "(",
        ]
        self.post = [")"]


class Identity(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents, single_arg=True)
        self.op = []

    def math_img_list(self):
        return self.arg1.math_img_list()

    @property
    def value(self):
        return self.arg1.value


class Subtensor(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents)
        subset = int(self.arg2.value)
        self.op = [rf"_[{subset}]"]
        #TODO: Consider changing how Subtensor MathOP handles arg2.
        self.arg2 = None


class Reciprocal(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents, single_arg=True)
        self.op = ["1/"]


class InverseLogit(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents, single_arg=True)
        self.pre = [r"\mathrm{invlogit}", "("]
        self.post = [")"]


class Dot(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents)
        self.op = ["⋅"]


class UnknownOp(MathOP):
    def __init__(self, apply_parents):
        super().__init__(apply_parents, single_arg=True)
        try:
            self.arg2 = recursive_f_search(apply_parents[1])
        except:
            pass
        self.pre = ["("]
        self.op = ["?"]
        self.post = [")"]


# TODO: Add all math operations
mathop_dict = {
    # Old Pytensor Syntax: retained for now, might no longer be useful. 
    "Elemwise{add,": {"op": Add},
    "Elemwise{mul,": {"op": Multiply},
    "Elemwise{true_div,": {"op": Divide},
    "Elemwise{pow,": {"op": Power},
    "Elemwise{exp,": {"op": Exp},
    "Elemwise{identity}": {"op": Identity},
    "InplaceDimShuffle": {"op": Identity},
    "dot": {"op": Dot},
    "Sum{": {"op": Identity},
    "Subtensor{": {"op": SubSet},
    "Elemwise{log10,": {"op": Logarithm10},
    "Elemwise{log2,": {"op": Logarithm2},
    "Elemwise{log,": {"op": Logarithm},
    "AdvancedSubtensor": {"op": SubSet},
    "Elemwise{sigmoid,": {"op": InverseLogit},

    # New Pytensor Syntax:

    "Add": {"op": Add},
    "Mul": {"op": Multiply},
    "Sub": {"op": Subtract},
    "True_div": {"op": Divide},
    "Int_div": {"op": Divide},
    "Reciprocal": {"op": Reciprocal},
    "Pow": {"op": Power},
    "Exp": {"op": Exp},
    "Identity": {"op": Identity},
    "InplaceDimShuffle": {"op": Identity},
    "dot": {"op": Dot},
    "Sum{": {"op": Identity},
    "Subtensor{": {"op": SubSet},
    "Log10": {"op": Logarithm10},
    "Log2": {"op": Logarithm2}, 
    "Log.": {"op": Logarithm}, # Period is there to differentiate from other log bases 
                               # It seems as though all OP names are followed by periods.
    "AdvancedSubtensor": {"op": SubSet},
    "Sigmoid": {"op": InverseLogit},
    "ExpandDims": {"op": Identity},


    "UNKNOWN": {"op": UnknownOp},
}


# endregion


### Other OPs that can result from a recursive f search:


class OtherOP:
    OP_ID = itertools.count()

    def __init__(self):
        self.name: str
        self.op_id = "other" + str(OtherOP.OP_ID.__next__())

    def __repr__(self):
        return str(self.name)

    def __str__(self):
        return self.__repr__()

    def op_to_graph(self, graph: AGraph, f_node_name):
        if graph.has_node(self.name):
            graph.add_edge(self.name, f_node_name, key=self.op_id)

    def self_img(self):
        raise NotImplementedError

    def math_img_list(self) -> list:
        return [self.self_img()]

    def latex_as_data(self):
        return string_to_latex(f"{DATA_LETTER}_{self.name}")

    # Replacing underscores with escaped underscores ensures
    # That no subsetting will happen in the string
    def latex_as_subset(self):
        return string_to_latex(self.name.replace("_", "\_"))

    def latex_as_variable(self):
        return string_to_latex(self.name)

    def latex_as_op(self):
        raise NotImplementedError


# region
class UnnamedDataOP(OtherOP):
    # This is a class attribute used to keep track of each instance of DataOp that has
    # been added to the graph; the effect will be purely cosmetic, but it will help differentiate
    # the various forms of data
    data_iterator = 1

    def __init__(self):
        super().__init__()
        self.name = f"{UnnamedDataOP.data_iterator}"
        UnnamedDataOP.data_iterator += 1

    def op_to_graph(self, graph, f_node_name):
        pass

    def self_img(self):
        mi = FC.fig_path(FC, symbols=self.latex_as_data(), op_id=self.op_id)
        if not isinstance(mi, MathImageFound):
            render_img_from_latex(self.latex_as_data(), mi.filepath)
        return mi


class NamedDataOP(OtherOP):
    def __init__(self, name):
        super().__init__()
        self.name = f"{name}"

    def self_img(self):
        mi = FC.fig_path(
            FC,
            edges_from=[str(self.name)],
            symbols=self.latex_as_data(),
            op_id=self.op_id,
        )
        if not isinstance(mi, MathImageFound):
            render_img_from_latex(self.latex_as_data(), mi.filepath)
        return mi


class NamedOP(OtherOP):
    def __init__(self, name):
        super().__init__()
        self.name = str(name)

    def self_img(self):
        mi = FC.fig_path(
            FC,
            edges_from=[str(self.name)],
            symbols=self.latex_as_variable(),
            op_id=self.op_id,
        )
        if not isinstance(mi, MathImageFound):
            render_img_from_latex(self.latex_as_variable(), mi.filepath)
        return mi


class ScalarOP(OtherOP):
    def __init__(self, value):
        super().__init__()
        self.value = value
        if float(self.value).is_integer():
            self.name = str(int(self.value))
        else:
            self.name = str(self.value)

    def __repr__(self):
        return str(self.value)

    def __str__(self):
        return self.__repr__()

    def op_to_graph(self, graph, f_node_name):
        pass

    def self_img(self):
        mi = FC.fig_path(
            FC,
            edges_from=[str(self.name)],
            symbols=self.latex_as_variable(),
            op_id=self.op_id,
        )
        if not isinstance(mi, MathImageFound):
            render_img_from_latex(str(self.value), mi.filepath)
        return mi


# endregion


def recursive_f_search(node: Variable) -> MathOP | ScalarOP | OtherOP:
    # Get the node's number of parents; we'll need to know it later.
    lenparents = len(node.get_parents())

    # If the node has a name, it's either named data or a named random variable.
    # If it has no parents, it's almost certainly data.
    # Otherwise, it's a random variable.
    try:
        if node.name and lenparents == 0:
            return NamedDataOP(str(node.name))

        if node.name and lenparents >= 0:
            return NamedOP(str(node.name))

    except AttributeError:
        pass

    # If the node has no name and no parents, we can be certain that it's
    # unnamed data -- we merely need to decide if it's a scalar (in which
    # case it's almost certainly a parameter), or if it's unnamed data

    if lenparents == 0:
        try:
            return ScalarOP(float(node.eval()))
        except TypeError:
            return UnnamedDataOP()

    mathop_type = None

    for type in mathop_dict.keys():
        if str(node).startswith(type):
            mathop_type = mathop_dict[type]
            continue

    if not mathop_type:
        print(f"Pykrusch didn't recognize {str(node)} as a valid MathOP type")
        mathop_type = mathop_dict["UNKNOWN"]

    mathop_op = mathop_type["op"]

    expression = mathop_op(node.owner.get_parents())

    return expression


def start_recursive_f_search(node: Variable) -> MathOP | OtherOP:
    if str(node.owner).startswith("Elemwise{identity") or str(node.owner).startswith("Identity"):
        return Identity(node.owner.get_parents())
    else:
        raise Exception(
            f"""
            Pykrusch doesn't know how to handle deterministic
            variables that don't begin with an Identity operation.

            This node was named: {str(node.owner)}
            """
        )


def specify_f_nodes(treelist) -> dict[str, list]:
    return [DeterministicVar(node) for node in treelist]


def add_f_nodes(nodes: list[DeterministicVar], graph: AGraph):
    for fnode in nodes:
        fnode.f_add_self_to_graph(graph)



import pymc as pm
import numpy as np

x = np.random.normal(loc=0, scale=1, size=10) # Data Placeholder
y = np.random.normal(loc=2*x, scale=1, size=10) # Data Placeholder


with pm.Model() as model:

	# Priors
	α = pm.Normal("α", 1, 2)
	β = pm.Normal("β", -1, 2)
	σ = pm.Exponential("σ", 3)

	# Linear Model
	μ = pm.Deterministic("μ", α + β*x)
	
	# Likelihood
	y_ = pm.Normal("y", μ, σ, observed=y)



with model:
	trace = pm.sample()


if __name__ == "__main__":

	from pykrusch import krusch
	krusch(model, outname="../img/simple_model_posterior.png", posterior_trace=trace)


from pygraphviz import AGraph
from pykrusch.config import NODE_COLOUR_GV
from pykrusch.dist import Dist, get_rv_dist
from pykrusch.param import (
    NamedParameter,
    UnknownParameter,
    NumericalParameter,
    DataParameter,
    NamedData,
    ComponentsParameter,
)
from pytensor.graph.basic import Variable
from pykrusch.graphviz.pgvHTML import make_rv_html, make_data_html


# THIS IS HOW YOU STOP CIRCULAR IMPORTS FOR TYPE CHECKING:
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pykrusch.dist import Dist


# Random Variable Handling

# Strategy thus far:
# Step 1: "Mix down" the entire PYMC graph as a pytensor function with no inputs (this preserves the parameters for each of the priors/hyperpriors)
# Step 2: for random variable in the graph, get the owner:
# - As far as I can tell, random variables will always have some kind of `_rv` as the sole owner
# - I can use this to determine the distribution shape (I think), which can then be used to follow important parameters
# Step 3: Follow the origin of the parameters until you come across a named node or an input (defined as a node with no parents, I think)
# - Sometimes, this will be a named node, in which case it's sufficient to merely find it and report it
# - Other times, this will be a
#

# for `normal_rv`, subscript 3 = location (mu), subscript 4 = scale (sigma)
# Only named nodes have a '.name' attribute

# When searching the tree, we're going to terminate at:
# 1. Another Named Node
# 2. A scalar
# 3. Some kinda mathematical operation


class RandomVar:
    def __init__(
        self,
        node,
        plot_posterior,
        posterior_data,
    ):
        self.name: str = node.name
        self.dist: Dist = get_rv_dist(node)
        self.dist.plot_posterior = plot_posterior
        self.dist.posterior_data = posterior_data

        self.shape = "plaintext"
        self.color = NODE_COLOUR_GV
        self.style = "filled"

    def rv_add_self_to_graph(self, graph: AGraph):
        # Add the node plus attributes
        graph.add_node(
            self.name,
            shape=self.shape,
            color=self.color,
            style=self.style,
            label=make_rv_html(self, graph),
        )
        for param in self.dist.params:
            self.connect_param(param=param, graph=graph)

    def connect_param(self, param, graph):
        if isinstance(param, NumericalParameter):
            pass
        elif isinstance(param, NamedParameter):
            graph.add_edge(
                param.name,
                self.name,
                tailport="s",
                headport=str(param.slot) + ":n",
                key=str(param.slot),
            )
        elif isinstance(param, NamedData):
            if not graph.has_node(param.name):
                graph.add_node(
                    param.name,
                    shape="plaintext",
                    color=NODE_COLOUR_GV,
                    style="filled",
                    label=make_data_html(param, graph),
                )
            graph.add_edge(
                param.name,
                self.name,
                tailport="s",
                headport=str(param.slot) + ":n",
                key=str(param.slot),
            )

        elif isinstance(param, ComponentsParameter):
            for component in param.flat_components:
                self.connect_param(component, graph)

        elif isinstance(param, UnknownParameter):
            pass

        elif isinstance(param, DataParameter):
            pass

        else:
            raise Exception(
                f"PyKrusch doesn't know how to handle treating {param} as a {self.dist} Random Variable"
            )


def specify_rv_nodes(
    treelist: list[Variable],
    plot_posterior: bool = False,
    posterior_data=None,
) -> list[RandomVar]:
    return [RandomVar(node, plot_posterior, posterior_data) for node in treelist]


def add_rv_nodes(rv_nodes: list[RandomVar], graph: AGraph):
    for rv in rv_nodes:
        rv.rv_add_self_to_graph(graph)



import matplotlib.pyplot as plt
from pykrusch.config import PARAM_CONVERSION
from re import split
from pykrusch.config import *


def string_to_latex(math_str) -> str:
    parts = []
    # Need to make sure that escaped '_' symbols aren't split:

    if LATEX_NAMES:
        for str_part in split(r"(?<!\\)_", math_str):
            if str_part in PARAM_CONVERSION:
                parts.append(PARAM_CONVERSION[str_part])
            else:
                parts.append(str_part)

        if len(parts) == 1:
            latex_str = rf"{parts[0]}"
        else:
            the_rest = "\_".join(parts[1:])
            latex_str = rf"{parts[0]}_{{{the_rest}}}"

        return latex_str

    else:
        return math_str.replace("\\", "").replace("{", "").replace("}", "")


def render_img_from_latex(latex_str: str, filepath):
    plt.style.use("default")

    plt.rcParams["mathtext.fontset"] = "cm"

    if LATEX_NAMES:
        mpl_string = rf"${latex_str}$"
    else:
        mpl_string = rf"{latex_str}".replace("\\", "").replace("{", "").replace("}", "")

    fig = plt.figure()
    fig.patch.set_facecolor(NODE_COLOUR_MPL)
    text = fig.text(0.5, 0.5, s=rf"{mpl_string}", ha="center", va="center", size=15)
    plt.savefig(
        filepath,
        bbox_inches="tight",
        dpi=DPI,
        pad_inches=0,
    )
    plt.close()



from __future__ import annotations
from pykrusch.param import (
    NumericalParameter,
    NamedParameter,
    UnknownParameter,
)
from pykrusch.param import recursive_rv_search
from pykrusch.figureControl import FC, MathImage, MathImageFound
from pykrusch.graphviz.distrogramsMPL import (
    plot_distrogram,
    plot_unknown_distrogram,
    plot_discrete_distrogram,
)
from pykrusch.graphviz.latexText import string_to_latex, render_img_from_latex
from typing import TYPE_CHECKING
from collections import namedtuple
import scipy.stats as sps
import numpy as np
import pymc as pm

np.seterr(divide="ignore")

if TYPE_CHECKING:
    from pykrusch.param import Parameter
    from pykrusch.parsePymc.randomVariable import RandomVar
    from arviz import InferenceData

_ = 0

partup = namedtuple(
    "Param",
    [
        "greek_name",
        "pos",
        "meaning",
        "op",
    ],
    defaults=[None],
)


def reciprocal(n):
    return 1.0 / n


class Dist:
    def __init__(self, owner, varname=""):
        # These attributes are to be overwritten by the specific classes
        self.needed_params: list[partup]
        self.type: str
        self.distimage: str
        self.fz: sps.rv_continuous | sps.rv_discrete
        self.numerical: bool = False
        self.plot_posterior: bool = False
        self.posterior_data: InferenceData | None = None
        self.scipy = True

        # These attributes are NOT to be overwritten
        self.owner = owner
        self.params: list[Parameter] = self.get_params()
        self.varname = varname
        self.num_params: int = len(self.params)

    def get_params(self):
        parameters = []
        for needed_param in self.needed_params:
            result = recursive_rv_search(self.owner.get_parents()[needed_param.pos])
            result.give_slot(needed_param.pos)
            result.give_greek(needed_param.greek_name)
            result.give_meaning(needed_param.meaning)

            # Apply the special operation to numerical values that get fed through
            if needed_param.op and isinstance(result, NumericalParameter):
                result.value = needed_param.op(result.value)

            parameters.append(result)

        return parameters

    @property
    def dist_name_latex(self):
        mi: MathImage = FC.fig_path(FC, symbols=self.varname)
        if not isinstance(mi, MathImageFound):
            render_img_from_latex(string_to_latex(self.varname), mi.filepath)
        return mi.filepath

    ### TODO: include something here to change between pdf and pmf
    ### depending on whether the distribution is continuous or discrete

    @property
    def distimage(self):
        mi_distrogram: MathImage = FC.fig_path(FC)

        plot_distrogram(
            dist=self,
            mi=mi_distrogram,
            plot_posterior=self.plot_posterior,
            posterior_data=self.posterior_data,
        )
        return mi_distrogram.filepath

    def ppf(self, x):
        return self.fz.ppf(x)

    def pdf(self, x):
        return self.fz.pdf(x)

    @property
    def xmin(self):
        return self.ppf(0.01) - self.xbuffer

    @property
    def xmax(self):
        return self.ppf(0.99) + self.xbuffer

    @property
    def xbuffer(self):
        range = self.ppf(0.99) - self.ppf(0.01)
        return range * 0.1

    def __repr__(self):
        return str(self.type)

    def __str__(self):
        return self.__repr__()


class UnknownDist(Dist):
    def __init__(self, owner, varname=""):
        self.type = "unknown"
        self.numerical = False
        self.owner = owner
        self.params: list[Parameter] = self.get_params()
        self.varname = varname

        self.num_params = 1

    # need to make sure that all of these are

    def get_params(self):
        parameters = []
        for param_slot in self.owner.get_parents()[3:]:
            result = recursive_rv_search(param_slot)
            result.greek_name = "?"
            result.meaning = " "

            if isinstance(result, NamedParameter):
                parameters.append(result)

        if len(parameters) == 0:
            result = UnknownParameter()
            result.greek_name = "?"
            result.meaning = " "
            result.slot = 1

            parameters.append(result)

        return parameters

    @property
    def distimage(self):
        mi_distrogram: MathImage = FC.fig_path(FC)

        plot_unknown_distrogram(dist=self, mi=mi_distrogram)
        return mi_distrogram.filepath


class Normal(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Normal"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("sigma", 4, "scale"),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.fz = sps.norm(
                loc=self.params[0].value,
                scale=self.params[1].value,
            )
            self.numerical = True
        else:
            self.fz = sps.norm()


class Exponential(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Exponential"
        self.needed_params = [
            partup("lambda", 3, "rate"),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.fz = sps.expon(scale=1 / self.params[0].value)
            self.numerical = True
        else:
            self.fz = sps.expon(1)


class Gamma(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Gamma"
        self.needed_params = [
            partup("alpha", 3, "shape"),
            partup("beta", 4, "rate", reciprocal),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.fz = sps.gamma(
                a=self.params[0].value,
                scale=1 / self.params[1].value,
            )
            self.numerical = True
        else:
            self.fz = sps.gamma(2, 2)


class Weibull(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Weibull"
        self.needed_params = [partup("alpha", 3, "shape"), partup("beta", 4, "scale")]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.numerical = True
        else:
            self.fz = sps.weibull_min(2, 2)

    def pdf(self, x):
        if self.numerical:
            return sps.weibull_min.pdf(
                x,
                c=self.params[0].value,
                scale=self.params[1].value,
            )
        else:
            return self.fz(x)

    def ppf(self, x):
        if self.numerical:
            return sps.weibull_min.ppf(
                x,
                c=self.params[0].value,
                scale=self.params[1].value,
            )
        else:
            return self.fz(x)


class Uniform(Dist):
    def __init__(self, owner, varname=""):
        self.type = "uniform"
        self.needed_params = [partup("", 3, "lower"), partup("", 4, "upper")]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            lower = self.params[0].value
            upper = self.params[1].value
            scale = upper - lower
            self.fz = sps.uniform(loc=lower, scale=scale)
            self.numerical = True
        else:
            self.fz = sps.uniform()


class TruncatedNormal(Dist):
    def __init__(self, owner, varname=""):
        self.type = "truncated normal"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("sigma", 4, "scale"),
            partup("", 5, "lower"),
            partup("", 6, "upper"),
        ]
        super().__init__(owner, varname)

        fz = sps.truncnorm

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.fz = fz(
                loc=self.params[0].value,
                scale=self.params[1].value,
                a=self.params[2].value,
                b=self.params[3].value,
            )
            self.numerical = True
        else:
            self.fz = fz(
                loc=0,
                scale=1,
                a=-1,
                b=1,
            )


class Beta(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Beta"
        self.needed_params = [
            partup("alpha", 3, "shape"),
            partup("beta", 4, "shape"),
        ]
        super().__init__(owner, varname)

        fz = sps.beta

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.fz = fz(
                a=self.params[0].value,
                b=self.params[1].value,
            )
            self.numerical = True
        else:
            self.fz = fz(
                a=3,
                b=5,
            )

    @property
    def xmin(self):
        return 0

    @property
    def xmax(self):
        return 1


class Kumaraswamy(Dist):
    def __init__(self, owner, varname=""):
        self.type = "kumaraswamy"
        self.needed_params = [
            partup("alpha", 3, "shape"),
            partup("beta", 4, "shape"),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.a = self.params[0].value
            self.b = self.params[1].value
            self.numerical = True
        else:
            self.a = 3
            self.b = 5

    def pdf(self, x):
        a = self.a
        b = self.b

        return a * b * (x ** (a - 1)) * ((1 - (x**a)) ** (b - 1))

    @property
    def xmin(self):
        return 0

    @property
    def xmax(self):
        return 1


class Laplace(Dist):
    def __init__(self, owner, varname=""):
        self.type = "laplace"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("b", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.b = self.params[1].value
            #             _ = self.params[2].value,
            #             _ = self.params[3].value,

            self.numerical = True
        else:
            self.mu = 3
            self.b = 5
        #                 _ = 7,
        #                 _ = 9,

        self.fz: sps.rv_continuous = sps.laplace(
            scale=self.b,
            loc=self.mu
            # _=self._,
            # _=self._,
        )

    def ppf(self, x):
        return self.fz.ppf(x)

    def pdf(self, x):
        return self.fz.pdf(x)


class StudentT(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Student T"
        self.needed_params = [
            partup("nu", 3, "d.o.f."),
            partup("mu", 4, "location"),
            partup("sigma", 5, "scale"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.nu = self.params[0].value
            self.mu = self.params[1].value
            self.sigma = self.params[2].value
            #             _ = self.params[3].value,

            self.numerical = True
        else:
            self.nu = 3
            self.mu = 5
            self.sigma = 7
        #                 _ = 9,

        self.fz: sps.rv_continuous = sps.t

    def ppf(self, x):
        return self.fz.ppf(x, df=self.nu, loc=self.mu, scale=self.sigma)

    def pdf(self, x):
        return self.fz.pdf(x, df=self.nu, loc=self.mu, scale=self.sigma)


class HalfStudent(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Half Student"
        self.needed_params = [
            partup("nu", 3, "d.o.f."),
            partup("sigma", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)
        self.scipy = False

        self.nu: float
        self.sigma: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.nu = self.params[0].value
            self.sigma = self.params[1].value
            self.numerical = True

        else:
            self.nu = 3
            self.sigma = 5

    def pdf(self, x):
        fz = pm.HalfStudentT.dist(nu=self.nu, sigma=self.sigma)
        out = pm.logp(fz, x).eval()
        return np.e ** (out)

    @property
    def xmin(self):
        return -0.1

    @property
    def xmax(self):
        return 5


class Cauchy(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Cauchy"
        self.needed_params = [
            partup("alpha", 3, "location"),
            partup("beta", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.a = self.params[0].value
            self.b = self.params[1].value
            #             self._ = self.params[2].value,
            #             self._ = self.params[3].value,

            self.numerical = True
        else:
            self.a = 0
            self.b = 1
        #                 self._ = 7,
        #                 self._ = 9,

        self.fz: sps.rv_continuous = sps.cauchy

    def ppf(self, x):
        return self.fz.ppf(x, loc=self.a, scale=self.b)

    def pdf(self, x):
        return self.fz.pdf(x, loc=self.a, scale=self.b)


class HalfCauchy(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Half Cauchy"
        self.needed_params = [
            partup("beta", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.beta: float
        self.__: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.beta = self.params[0].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.beta = 3
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.halfcauchy

    def ppf(self, x):
        result = self.fz.ppf(x, scale=self.beta)
        return result

    def pdf(self, x):
        return self.fz.pdf(x, scale=self.beta)


class LogNormal(Dist):
    def __init__(self, owner, varname=""):
        self.type = "LogNormal"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("sigma", 4, "scale"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = (self.params[0].value,)
            self.sigma = (self.params[1].value,)
            #             self._ = self.params[3].value,

            self.numerical = True
        else:
            self.mu = (3,)
            self.sigma = (5,)
        #                 self._ = 9,

        self.fz: sps.rv_continuous = sps.lognorm

    def ppf(self, x):
        return self.fz.ppf(x, 1, loc=self.mu, scale=self.sigma)

    def pdf(self, x):
        return self.fz.pdf(x, 1, loc=self.mu, scale=self.sigma)

    @property
    def xmin(self):
        return self.ppf(0.1) - self.xbuffer

    @property
    def xmax(self):
        return self.ppf(0.9) + self.xbuffer

    @property
    def xbuffer(self):
        range = self.ppf(0.9) - self.ppf(0.1)
        return range * 0.1


class ChiSquared(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Chi-Squared"
        self.needed_params = [
            partup("nu", 3, "d.o.f."),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.nu = self.params[0].value
            #             self._ = self.params[2].value,
            #             self._ = self.params[3].value,

            self.numerical = True
        else:
            self.nu = 3
        #                 self._ = 7,
        #                 self._ = 9,

        self.fz: sps.rv_continuous = sps.chi2(
            df=self.nu,
            #            _=self._,
            #            _=self._,
        )


class HalfNormal(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Half Normal"
        self.needed_params = [
            partup("sigma", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.sigma: float
        self.__: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.sigma = self.params[0].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.sigma = 3
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.halfnorm

    def ppf(self, x):
        result = self.fz.ppf(x, scale=self.sigma)
        return result

    def pdf(self, x):
        return self.fz.pdf(x, scale=self.sigma)


class Wald(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Wald"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("lam", 4, "precision"),
            partup("alpha", 5, "shift"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.mu: float
        self.lam: float
        self.alpha: float
        # self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.lam = self.params[1].value
            self.alpha = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 3
            self.lam = 5
            self.alpha = 7

    #                 self.____ = 9

    def pdf(self, x):
        mu = self.mu
        lam = self.lam
        term1 = (lam / (2 * np.pi)) ** (1 / 2)
        term2 = x ** ((-3) / 2)
        term3 = (-lam / (2 * x)) * (((x - mu) / mu) ** 2)
        return term1 * term2 * (np.e ** (term3))

    @property
    def xmin(self):
        return 0

    @property
    def xmax(self):
        return 10


class Pareto(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Pareto"
        self.needed_params = [
            partup("alpha", 3, "shape"),
            partup("m", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.alpha: float
        self.m: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.alpha = self.params[0].value
            self.m = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.alpha = 3
            self.m = 5
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.pareto

    def ppf(self, x):
        return self.fz.ppf(x, b=self.alpha, scale=self.m)

    def pdf(self, x):
        return self.fz.pdf(x, b=self.alpha, scale=self.m)


class InverseGamma(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Inverse Gamma"
        self.needed_params = [
            partup("alpha", 3, "shape"),
            partup("beta", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.alpha: float
        self.beta: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.alpha = self.params[0].value
            self.beta = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.alpha = 3
            self.beta = 5
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.invgamma

    def ppf(self, x):
        return self.fz.ppf(x, a=self.alpha, scale=self.beta)

    def pdf(self, x):
        return self.fz.pdf(x, a=self.alpha, scale=self.beta)


class ExGaussian(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Exponential Gaussian"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("sigma", 4, "scale"),
            partup("nu", 5, "exponential"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.mu: float
        self.sigma: float
        self.nu: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.sigma = self.params[1].value
            self.nu = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 3
            self.sigma = 5
            self.nu = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.exponnorm

    def ppf(self, x):
        return self.fz.ppf(x, K=self.nu, loc=self.mu, scale=self.sigma)

    def pdf(self, x):
        return self.fz.pdf(x, K=self.nu, loc=self.mu, scale=self.sigma)


class VonMises(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Von Mises"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("kappa", 4, "concentration"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.mu: float
        self.kappa: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.kappa = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 3
            self.kappa = 5
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.vonmises

    def ppf(self, x):
        return self.fz.ppf(x, kappa=self.kappa, loc=self.mu)

    def pdf(self, x):
        return self.fz.pdf(x, kappa=self.kappa, loc=self.mu)


class SkewNormal(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Skew Normal"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("sigma", 4, "scale"),
            partup("alpha", 5, "skew"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.mu: float
        self.sigma: float
        self.alpha: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.sigma = self.params[1].value
            self.alpha = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 3
            self.sigma = 5
            self.alpha = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.skewnorm

    def ppf(self, x):
        return self.fz.ppf(x, a=self.alpha, loc=self.mu, scale=self.sigma)

    def pdf(self, x):
        return self.fz.pdf(x, a=self.alpha, loc=self.mu, scale=self.sigma)


class Triangular(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Triangular"
        self.needed_params = [
            partup("", 3, "lower"),
            partup("c", 4, "mode"),
            partup("", 5, "upper"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.lower: float
        self.c: float
        self.upper: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.lower = self.params[0].value
            self.c = self.params[1].value
            self.upper = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.lower = 3
            self.c = 5
            self.upper = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.triang

    def ppf(self, x):
        c = (self.c - self.lower) / (self.upper - self.lower)
        upper = self.upper - self.lower
        result = self.fz.ppf(x, c=c, loc=self.lower, scale=upper)
        return result

    def pdf(self, x):
        c = (self.c - self.lower) / (self.upper - self.lower)
        upper = self.upper - self.lower
        result = self.fz.pdf(x, c=c, loc=self.lower, scale=upper)
        return result


class Gumbel(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Gumbel"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("beta", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.mu: float
        self.beta: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.beta = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 3
            self.beta = 5
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.gumbel_r

    def ppf(self, x):
        return self.fz.ppf(x, loc=self.mu, scale=self.beta)

    def pdf(self, x):
        return self.fz.pdf(x, loc=self.mu, scale=self.beta)


class Logistic(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Logistic"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("beta", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.mu: float
        self.s: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.s = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 3
            self.s = 5
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.logistic

    def ppf(self, x):
        return self.fz.ppf(x, loc=self.mu, scale=self.s)

    def pdf(self, x):
        return self.fz.pdf(x, loc=self.mu, scale=self.s)


class LogitNormal(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Logit Normal"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("sigma", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)
        self.scipy = False

        self.mu: float
        self.sigma: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.sigma = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 1
            self.sigma = 2

    #                 self.___ = 7
    #                 self.____ = 9

    def pdf(self, x):
        fz = pm.LogitNormal.dist(mu=self.mu, sigma=self.sigma)
        out = pm.logp(fz, x).eval()
        return np.e ** (out)

    @property
    def xmin(self):
        return -0.1

    @property
    def xmax(self):
        return 1.1

    @property
    def distimage(self):
        mi_distrogram: MathImage = FC.fig_path(FC)

        plot_unknown_distrogram(dist=self, mi=mi_distrogram)
        return mi_distrogram.filepath


class Rice(Dist):
    def __init__(self, owner, varname=""):
        self.type = "rice"
        self.needed_params = [
            partup("b", 3, "shape"),
            partup("sigma", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.b: float
        self.sigma: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.b = self.params[0].value
            self.sigma = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.b = 3
            self.sigma = 5
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.rice

    def ppf(self, x):
        return self.fz.ppf(x, b=self.b, scale=self.sigma)

    def pdf(self, x):
        return self.fz.pdf(x, b=self.b, scale=self.sigma)


class Moyal(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Moyal"
        self.needed_params = [
            partup("mu", 3, "location"),
            partup("sigma", 4, "scale"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.mu: float
        self.sigma: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            self.sigma = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 3
            self.sigma = 5
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.moyal

    def ppf(self, x):
        return self.fz.ppf(x, loc=self.mu, scale=self.sigma)

    def pdf(self, x):
        return self.fz.pdf(x, loc=self.mu, scale=self.sigma)


class AsymmetricLaplace(Dist):
    def __init__(self, owner, varname=""):
        self.type = "Asymmetric Laplace"
        self.needed_params = [
            partup("kappa", 4, "symmetry"),
            partup("mu", 5, "location"),
            partup("b", 3, "scale"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.kappa: float
        self.mu: float
        self.b: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.kappa = self.params[0].value
            self.mu = self.params[1].value
            self.b = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.kappa = 3
            self.mu = 5
            self.b = 7
        #                 self.____ = 9

        self.fz: sps.rv_continuous = sps.laplace_asymmetric

    def ppf(self, x):
        return self.fz.ppf(x, kappa=self.kappa, loc=self.mu, scale=self.b)

    def pdf(self, x):
        return self.fz.pdf(x, kappa=self.kappa, loc=self.mu, scale=self.b)


class DiscreteDist(Dist):
    def pmf(self, x):
        return self.fz.pmf(x)

    @property
    def distimage(self):
        mi_distrogram: MathImage = FC.fig_path(FC)

        plot_discrete_distrogram(
            dist=self,
            mi=mi_distrogram,
            plot_posterior=self.plot_posterior,
            posterior_data=self.posterior_data,
        )
        return mi_distrogram.filepath

    @property
    def xmin(self):
        return self.ppf(0.01)

    @property
    def xmax(self):
        return self.ppf(0.99)


class DiscreteUniform(DiscreteDist):
    def __init__(self, owner, varname=""):
        self.type = "Discrete Uniform"
        self.needed_params = [
            partup("", 3, "lower"),
            partup("", 4, "upper"),
        ]
        super().__init__(owner, varname)

        self.lower: float
        self.upper: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.lower = self.params[0].value
            self.upper = self.params[1].value
            self.numerical = True

        else:
            self.lower = 3
            self.upper = 5

        self.fz: sps.rv_discrete = sps.randint

    def ppf(self, x):
        return self.fz.ppf(x, low=self.lower, high=self.upper)

    def pmf(self, x):
        results = self.fz.pmf(x, low=self.lower, high=self.upper)
        return results

    @property
    def xmin(self):
        return self.lower

    @property
    def xmax(self):
        return self.upper


class Binomial(DiscreteDist):
    def __init__(self, owner, varname=""):
        self.type = "Binomial"
        self.needed_params = [
            partup("n", 3, "trials"),
            partup("p", 4, "P(success)"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.n: float
        self.p: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.n = self.params[0].value
            self.p = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.n = 10
            self.p = 0.3
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_discrete = sps.binom

    def ppf(self, x):
        return self.fz.ppf(x, n=self.n, p=self.p)

    def pmf(self, x):
        return self.fz.pmf(x, n=self.n, p=self.p)

    @property
    def xmin(self):
        return self.ppf(0.01)

    @property
    def xmax(self):
        return self.ppf(0.99)


class BetaBinomial(DiscreteDist):
    def __init__(self, owner, varname=""):
        self.type = "Beta Binomial"
        self.needed_params = [
            partup("n", 3, "trials"),
            partup("alpha", 4, "shape"),
            partup("beta", 5, "shape"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.n: float
        self.alpha: float
        self.beta: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.n = self.params[0].value
            self.alpha = self.params[1].value
            self.beta = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.n = 3
            self.alpha = 0.5
            self.beta = 7
        #                 self.____ = 9

        self.fz: sps.rv_discrete = sps.betabinom

    def ppf(self, x):
        return self.fz.ppf(x, n=self.n, a=self.alpha, b=self.beta)

    def pmf(self, x):
        return self.fz.pmf(x, n=self.n, a=self.alpha, b=self.beta)


class Bernoulli(DiscreteDist):
    def __init__(self, owner, varname=""):
        self.type = "Bernoulli"
        self.needed_params = [
            partup("p", 3, "probability"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.p: float
        self.__: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.p = self.params[0].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.p = 0.7
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_discrete = sps.bernoulli

    def ppf(self, x):
        return self.fz.ppf(x, p=self.p)

    def pmf(self, x):
        return self.fz.pmf(x, p=self.p)


class Poisson(DiscreteDist):
    def __init__(self, owner, varname=""):
        self.type = "Poisson"
        self.needed_params = [
            partup("mu", 3, "occurrences"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.mu: float
        self.__: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.mu = self.params[0].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.mu = 3
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_discrete = sps.poisson

    def ppf(self, x):
        return self.fz.ppf(x, mu=self.mu)

    def pmf(self, x):
        return self.fz.pmf(x, mu=self.mu)


class NegativeBinomial(DiscreteDist):
    def __init__(self, owner, varname=""):
        self.type = "Negative Binomial"
        self.needed_params = [
            partup("n", 3, "successes"),
            partup("p", 4, "probability"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.n: float
        self.p: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.n = self.params[0].value
            self.p = self.params[1].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.n = 3
            self.p = 0.5
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_discrete = sps.nbinom

    def ppf(self, x):
        return self.fz.ppf(x, n=self.n, p=self.p)

    def pmf(self, x):
        return self.fz.pmf(x, n=self.n, p=self.p)


class Geometric(DiscreteDist):
    def __init__(self, owner, varname=""):
        self.type = "Geometric"
        self.needed_params = [
            partup("p", 3, "probability"),
            #             partup("", 5, ""),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.p: float
        self.__: float
        self.___: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.p = self.params[0].value
            #             self.___ = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.p = 0.4
        #                 self.___ = 7
        #                 self.____ = 9

        self.fz: sps.rv_discrete = sps.geom

    def ppf(self, x):
        return self.fz.ppf(x, p=self.p)

    def pmf(self, x):
        return self.fz.pmf(x, p=self.p)


class Hypergeometric(DiscreteDist):
    def __init__(self, owner, varname=""):
        self.type = "Hypergeometric"
        self.needed_params = [
            partup("N", 3, "population"),
            partup("k", 4, "successes"),
            partup("n", 5, "samples"),
            #             partup("", 6, ""),
        ]
        super().__init__(owner, varname)

        self.N: float
        self.k: float
        self.n: float
        self.____: float

        if all(isinstance(p, NumericalParameter) for p in self.params):
            self.N = self.params[0].value + self.params[1].value
            self.k = self.params[0].value
            self.n = self.params[2].value
            #             self.____ = self.params[3].value

            self.numerical = True
        else:
            self.N = 25
            self.k = 5
            self.n = 15
        #                 self.____ = 9

        self.fz: sps.rv_discrete = sps.hypergeom

    def ppf(self, x):
        return self.fz.ppf(x, M=self.N, n=self.k, N=self.n)

    def pmf(self, x):
        return self.fz.pmf(x, M=self.N, n=self.k, N=self.n)


dist_dict: dict = {
    "UNKNOWN": UnknownDist,
    "exponential_rv": Exponential,
    "normal_rv": Normal,
    "gamma_rv": Gamma,
    "weibull_rv": Weibull,
    "uniform_rv": Uniform,
    "truncated_normal_rv": TruncatedNormal,
    "beta_rv": Beta,
    "kumaraswamy_rv": Kumaraswamy,
    "laplace_rv": Laplace,
    "studentt_rv": StudentT,
    "cauchy_rv": Cauchy,
    "halfcauchy_rv": HalfCauchy,
    "lognormal_rv": LogNormal,
    "chisquare_rv": ChiSquared,
    "halfnormal_rv": HalfNormal,
    "wald_rv": Wald,
    "pareto_rv": Pareto,
    "invgamma_rv": InverseGamma,
    "exgaussian_rv": ExGaussian,
    "vonmises_rv": VonMises,
    "skewnormal_rv": SkewNormal,
    "triangular_rv": Triangular,
    "halfstudentt_rv": HalfStudent,
    "gumbel_rv": Gumbel,
    "logistic_rv": Logistic,
    "logit_normal_rv": LogitNormal,
    "rice_rv": Rice,
    "moyal_rv": Moyal,
    "asymmetriclaplace_rv": AsymmetricLaplace,
    "binomial_rv": Binomial,
    "beta_binomial_rv": BetaBinomial,
    "bernoulli_rv": Bernoulli,
    "poisson_rv": Poisson,
    "nbinom_rv": NegativeBinomial,
    "discrete_uniform_rv": DiscreteUniform,
    "geometric_rv": Geometric,
    "hypergeometric_rv": Hypergeometric,
}


def get_rv_dist(node: RandomVar) -> Dist:
    owner = node.owner
    dist_type = str(owner).split("{")[0]

    # Get the distribution type from the dictionary thereof:
    try:
        return dist_dict[dist_type](owner, node.name)
    except KeyError:
        print(f"PyKrusch doesn't recognize {dist_type}")
        unknown_dist = dist_dict["UNKNOWN"](owner, node.name)
        unknown_dist.type = f"{dist_type}"
        if unknown_dist.type.endswith("_rv"):
            unknown_dist.type = unknown_dist.type.rpartition("_rv")[0]
        return unknown_dist



from __future__ import annotations
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats
from typing import TYPE_CHECKING
from pykrusch.config import *
from matplotlib.ticker import MaxNLocator

if TYPE_CHECKING:
    from pykrusch.figureControl import MathImage
    from pykrusch.dist import Dist, DiscreteDist
    from arviz.data import InferenceData


def plot_distrogram(
    dist: Dist,
    mi: MathImage,
    plot_posterior: bool = False,
    posterior_data: InferenceData | None = None,
):
    plt.style.use("default")
    xmin = dist.xmin
    xmax = dist.xmax
    ls = np.linspace(xmin, xmax, LINSPACE_N)
    plt.plot(ls, dist.pdf(ls), linewidth=DISTROGRAM_LINE_WIDTH)

    if plot_posterior and dist.numerical and dist.scipy:
        data_vars = list(posterior_data.posterior.data_vars)

        if dist.varname in data_vars:
            vals = posterior_data.posterior[dist.varname].values
            combined_trace = np.concatenate(vals)

            if combined_trace.ndim == 1:
                range_buffer = abs(max(combined_trace) - min(combined_trace)) * 0.1

                pxmin = min(combined_trace) - range_buffer
                pxmax = max(combined_trace) + range_buffer

                pls = np.linspace(pxmin, pxmax, LINSPACE_N)

                kde = stats.gaussian_kde(combined_trace)

                plt.plot(pls, kde.evaluate(pls), linewidth=DISTROGRAM_LINE_WIDTH)

            else:
                print(f"{dist.varname} did not have exactly one dimension")

        else:
            print(f"{dist.varname} not found in posterior data variables!")

    ax = plt.gca()
    ax.get_yaxis().set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)

    ax.tick_params(axis="both", which="major", labelsize=15)

    if (xmin - 1) < 0 and 0 < (xmax + 1) and dist.numerical and dist.scipy:
        plt.axvline(x=0, alpha=0.1, c="black", linewidth=DISTROGRAM_LINE_WIDTH)

    if not dist.numerical or not dist.scipy:
        plt.axis("off")

    plt.savefig(
        mi.filepath,
        bbox_inches="tight",
        dpi=DPI,
        pad_inches=0,
    )

    plt.close()


def plot_discrete_distrogram(
    dist: Dist | DiscreteDist,
    mi: MathImage,
    plot_posterior: bool = False,
    posterior_data: InferenceData | None = None,
):
    plt.style.use("default")
    xmin = dist.xmin
    xmax = dist.xmax + 1

    ar = np.arange(xmin, xmax)

    plt.bar(
        ar,
        dist.pmf(ar),
        linewidth=DISCRETE_LINE_WIDTH,
        color="None",
        edgecolor="tab:blue",
    )

    plt.bar([xmin], [-0.00001], color="white")

    if plot_posterior and dist.numerical and dist.scipy:
        data_vars = list(posterior_data.posterior.data_vars)

        if dist.varname in data_vars:
            combined_trace = np.concatenate(
                posterior_data.posterior[dist.varname].values
            )

            pxmin = min(combined_trace)
            pxmax = max(combined_trace) + 1
            par = np.arange(pxmin, pxmax)

            unique, counts = np.unique(combined_trace, return_counts=True)

            ardict = {}

            for i in par:
                ardict[i] = 0
                if i in unique:
                    ardict[i] += float(counts[np.where(unique == i)] / sum(counts))

            plt.bar(
                par,
                list(ardict.values()),
                linewidth=2,
                color="None",
                edgecolor="tab:orange",
                width=0.7,
            )

        else:
            print(f"{dist.varname} not found in posterior data variables!")

    ax = plt.gca()
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.get_yaxis().set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)

    ax.tick_params(axis="both", which="major", labelsize=15)

    if (xmin - 1) < 0 and 0 < (xmax + 1) and dist.numerical and dist.scipy:
        plt.axvline(x=0, alpha=0.1, c="black", linewidth=DISTROGRAM_LINE_WIDTH)

    if not dist.numerical or not dist.scipy:
        plt.axis("off")

    plt.savefig(
        mi.filepath,
        bbox_inches="tight",
        dpi=DPI,
        pad_inches=0,
    )

    plt.close()


def plot_unknown_distrogram(dist: Dist, mi: MathImage):
    plt.style.use("default")
    fig = plt.figure()

    ls = np.linspace(-3, 3, LINSPACE_N)
    plt.plot(
        ls,
        stats.norm.pdf(ls, loc=0, scale=1),
        linewidth=DISTROGRAM_LINE_WIDTH,
        c="white",
    )

    text = fig.text(0.5, 0.5, s="?", ha="center", va="center", size=70, c="tab:blue")
    ax = plt.gca()
    ax.get_yaxis().set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    plt.axis("off")
    plt.savefig(
        mi.filepath,
        bbox_inches="tight",
        dpi=DPI,
        pad_inches=0,
    )
    plt.close()



from os import mkdir, remove, rmdir
from pathlib import PosixPath
from pykrusch.config import *


class MathImage:
    def __init__(
        self,
        filepath: PosixPath,
        portnum: int,
        edges_from=[],
        op_id=None,
    ):
        self.filepath = filepath
        self.portnum = portnum
        self.edges_from: list = edges_from
        self.op_id = op_id


class MathImageFound(MathImage):
    def __init__(
        self,
        filepath: PosixPath,
        portnum: int,
        edges_from=[],
        op_id=None,
    ):
        self.filepath = filepath
        self.portnum = portnum
        self.edges_from: list = edges_from
        self.op_id = op_id


class FC:
    fignum = 0
    temppath = PosixPath(TEMP_IMG_DIR)

    symbol_dict = {}
    other_imgs = []

    def fig_init(fc):
        try:
            mkdir(fc.temppath)
        except FileExistsError:
            pass

    def fig_path(
        fc,
        edges_from=[],
        symbols="",
        op_id=None,
    ) -> MathImage | MathImageFound:
        fc.fignum += 1

        portnum = int(fc.fignum)

        if symbols and symbols in fc.symbol_dict:
            filepath = fc.symbol_dict[symbols]
            return MathImageFound(
                filepath=filepath, portnum=portnum, edges_from=edges_from, op_id=op_id
            )

        filepath = fc.temppath / f"{fc.fignum}.png"

        if symbols:
            fc.symbol_dict[symbols] = filepath
        else:
            fc.other_imgs.append(filepath)

        return MathImage(
            filepath=filepath, portnum=portnum, edges_from=edges_from, op_id=op_id
        )

    def fig_clean(fc):
        for f in fc.symbol_dict.values():
            try:
                remove(f)
            except FileNotFoundError:
                pass

        for f in fc.other_imgs:
            try:
                remove(f)
            except FileNotFoundError:
                pass

        try:
            rmdir(fc.temppath)
        except FileNotFoundError:
            pass



from pytensor.graph.basic import Apply
from pykrusch.config import PARAM_CONVERSION, PARAM_FONT_SIZE, PARAM_FONT_SIZE_NUMERICAL
from pykrusch.figureControl import MathImage, FC, MathImageFound
from pykrusch.graphviz.latexText import render_img_from_latex, string_to_latex


class Parameter:
    def __init__(self):
        self.name: str
        self.meaning: str
        self.printout: str
        self.type: str

        # This gets supplied in dist.py after instantiation
        self.greek_name: str
        self.slot: int

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"({self.type}, {self.printout})"

    def give_greek(self, greek):
        if greek in PARAM_CONVERSION:
            self.greek_name = PARAM_CONVERSION[greek]
        else:
            self.greek_name = greek

    def give_slot(self, slot):
        self.slot = slot

    def give_symbol(self):
        return self.greek_name, PARAM_FONT_SIZE

    def give_meaning(self, meaning):
        self.meaning = meaning


class NumericalParameter(Parameter):
    def __init__(self, value):
        self.name = ""
        self.value = value
        self.printout = str(value)
        self.type = "NumericalParameter"

    def give_symbol(self):
        if float(self.value).is_integer():
            numerical_arg = int(float(self.value))
        else:
            numerical_arg = float(self.value)

        if self.greek_name:
            return self.greek_name + f" = {numerical_arg}", PARAM_FONT_SIZE_NUMERICAL
        else:
            return f"{numerical_arg}", PARAM_FONT_SIZE_NUMERICAL


class NamedParameter(Parameter):
    def __init__(self, name):
        self.name = name
        self.printout = str(name)
        self.type = "NamedParameter"


class NamedData(Parameter):
    def __init__(self, name):
        self.name = name
        self.printout = str(name)
        self.type = "NamedData"

        mi: MathImage = FC.fig_path(FC, symbols=self.name)
        if not isinstance(mi, MathImageFound):
            render_img_from_latex(string_to_latex(self.name), mi.filepath)
        self.mi = mi


class DataParameter(Parameter):
    def __init__(self):
        self.name = ""
        # TODO: THIS NEEDS TO BE CHANGED
        self.printout = "DATA_PARAMETER"
        self.type = "DataParameter"


class UnknownParameter(Parameter):
    def __init__(self):
        self.name = ""
        self.printout = "UNKNOWN_PARAMETER"
        self.type = "UnknownParameter"


class ComponentsParameter(Parameter):
    def __init__(self, components):
        self.components = components
        self.name = [c.name for c in self.components]
        self.type = [c.type for c in self.components]
        self.printout = str(self.name)
        self.flat_components: set()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"({self.type}, {self.printout})"

    def give_slot(self, slot):
        self.slot = slot
        for component in self.flat_components:
            component: Parameter
            component.give_slot(slot)

    def give_greek(self, greek):
        self.greek_name = greek
        if greek in PARAM_CONVERSION:
            self.greek_name = PARAM_CONVERSION[greek]
        else:
            self.greek_name = greek

    def give_meaning(self, meaning):
        self.meaning = meaning
        for component in self.flat_components:
            component: Parameter
            component.give_meaning(meaning)

    def flatten_components(self):
        out_set = set()
        for component in self.components:
            if isinstance(component, ComponentsParameter):
                out_set = out_set.union(component.flatten_components())
            else:
                out_set.add(component)

        self.flat_components = out_set

        self.name = [c.name for c in self.flat_components]
        self.type = [c.type for c in self.flat_components]
        self.printout = str(self.name)
        return out_set


def recursive_rv_search(node) -> Parameter:
    # First, attempt to retrieve the node name, if there is one, return it

    parents = node.get_parents()

    try:
        if node.name:
            if hasattr(node, "data") or len(parents) == 0:
                return NamedData(node.name)
            return NamedParameter(node.name)
    except AttributeError:
        pass

    # If we make it past the try block above, it means the node
    # has no name.
    # If an unnamed node has 0 parents, it's either a parameter or data.
    # If it's a parameter, we should be able to retrieve it using
    # `node.data.item()`. If that fails, it's unnamed data.
    # So, let's get the parents:

    # Does the node have 0 parents? If so, it's either
    # a scalar (usually a parameter in a distribution)
    # or it's data. If it's a scalar, get it and return it.
    # If it's data, we simply call it 'x' for now.
    ###
    # TODO: Consider creating a better representation for data
    # Some kind of class, perhaps?
    ###
    if len(parents) == 0:
        try:
            return NumericalParameter(node.data.item())
        except ValueError:
            return DataParameter()

    # If the node has one parent, we go deeper.
    if len(parents) == 1:
        return recursive_rv_search(parents[0])

    # If the node is a math operation, has multiple parents,
    # and all of the parents are scalars, we'll try evaluating
    # to a scalar:
    if len(parents) == 2:
        try:
            parents[0].data.item()
            parents[1].data.item()
            # If the node type is an 'Apply' pytensor node,
            # We want to evaluate the apply node's CHILD.
            if type(node) == Apply:
                return NumericalParameter(float(node.outputs[0].eval()))
            else:
                return NumericalParameter(float(node.eval()))
        except NameError:
            pass
        except ValueError:
            pass
        except AttributeError:
            pass

    # If the above doesn't work, we check to see if the node is a
    # subtensor, which means that it's probably going to split a named
    # node into more than one component.

    # TODO: Improve handling here: currently, subtensor handling
    # simply treats them as invisible.

    if len(parents) == 2 and str(node).startswith("Subtensor"):
        return recursive_rv_search(parents[0])

        pass

    # If the node has multiple parents, we switch to a list of parameters/inputs.
    # This is just for debugging purposes.

    comp_list = []
    for parent in parents:
        result = recursive_rv_search(parent)
        comp_list.append(result)

    comp = ComponentsParameter(comp_list)
    comp.flatten_components()

    print(comp.flat_components)

    return comp



from __future__ import annotations
from typing import TYPE_CHECKING

import click
from pykrusch.config import *

from shutil import rmtree

if TYPE_CHECKING:
    from pygraphviz import AGraph
    from pymc import Model
    from arviz import InferenceData




# region
def posterior_data_check(
    MODEL: Model,
    posterior_trace: InferenceData | None = None,
    posterior_pickle: str | None = None,
    sample_posterior: bool = False,
):
    too_many = SystemExit(
        """More than one posterior source was passed. Please pass at most
        one of posterior_data, posterior_pickle, or sample_posterior=True."""
    )

    if posterior_trace and posterior_pickle:
        raise too_many
    if posterior_trace and sample_posterior:
        raise too_many
    if posterior_pickle and sample_posterior:
        raise too_many
    if posterior_trace:
        print(" ")
        print("Using supplied trace data.")
        print(" ")
        return True, posterior_trace
    if posterior_pickle:
        import pickle

        print(" ")
        print("Opening model trace pickle.")
        print(" ")
        with open(posterior_pickle, "rb") as stream:
            return True, pickle.load(stream)
    if sample_posterior:
        print(" ")
        print("Sampling pymc model.")
        print(" ")
        import pymc as pm
        with MODEL:
            return True, pm.sample()
        
    return False, None


def model_check(
    MODEL,
    model_file,
    model_name,
):
    if not MODEL and not model_file:
        raise SystemExit(
            """Please provide either MODEL (in the form of a pymc Model object), or model_file. 
Running pykrusch from the command line requires model_file."""
        )
    if MODEL and model_file:
        raise SystemExit(
            """Cannot pass both MODEL and model_file. Please provide exactly one of MODEL, model_file."""
        )
    if model_file and not model_name:
        raise SystemExit("""Use of model_file requires model_name.""")
    if MODEL:
        return MODEL
    if model_file and model_name:
        from importlib.util import spec_from_file_location, module_from_spec
        import sys

        spec = spec_from_file_location("_MODEL_FROM_FILE", model_file)
        mod = module_from_spec(spec)
        sys.modules["_MODEL_FROM_FILE"] = mod
        spec.loader.exec_module(mod)
        return mod.__dict__[model_name]


@click.command(
    context_settings=dict(help_option_names=["-h", "--help"]),
    help="""Visualize Bayesian model structure. 
Takes a Bayesian model specified in pymc and outputs an image portraying the dependency
structure formed by the model's random and deterministic variables, including prior
distribution (and, if specified, posterior distribution).""",
)
@click.argument(
    "model_file",
    type=click.Path(exists=True)
)
@click.option(
    "-n",
    "--model_name",
    default="model",
    help="Name of the model object in model_file. Defaults to 'model'.",
)
@click.option(
    "-p",
    "--posterior_pickle",
    default=None,
    help="Path to the .pkl file containing the sampling trace (only needed for plotting posterior).",
)
@click.option(
    "-s",
    "--sample_posterior",
    is_flag=True,
    default=False,
    help="Flag for sampling the posterior from the model. Can be slow (only needed if plotting posterior).",
)
@click.option(
    "-o",
    "--outname",
    default="krusch.png",
    help="Path of the output image. Defaults to 'krusch.png'.",
)
@click.option(
    "-r",
    "--retain_figs",
    is_flag=True,
    default=False,
    help="Flag for retaining the intermediary figures (defaults to False).",
)
def main(
    model_file: str | None = None,
    model_name: str | None = None,
    posterior_pickle: str | None = None,
    sample_posterior: bool = False,
    outname: str = "krusch.png",
    retain_figs=False,
):
    

    krusch_(
        pymc_model=None,
        model_file=model_file,
        model_name=model_name,
        posterior_trace=None,
        posterior_pickle=posterior_pickle,
        sample_posterior=sample_posterior,
        outname=outname,
        retain_figs=retain_figs,
    )


# endregion


def krusch(
        pymc_model: Model,
        posterior_trace: InferenceData|None = None,
        outname: str = "krusch.png"): 
    """Produces a visualization of a `pymc`-specified Bayesian Generalized Linear Model.

    Args:

        pymc_model (Model): 
            The `pymc` `Model` object to be visualized.

        posterior_trace (InferenceData | None, optional): 
            Optional: argument specifying the `arviz` `InferenceData` object containing the
            posterior information needed to plot the posterior distribution for each viable
            variable in the model. If supplied, `pykrusch` will plot the posterior to the best
            of its ability. If this argument is not supplied (or is `None`), the posterior 
            distributions will not be plotted.

        outname (str, optional): _description_. 
            Optional: filename of the resulting visualization. Defaults to "krusch.png".
    """
    krusch_(
        pymc_model=pymc_model,
        model_file=None,
        model_name=None,
        posterior_trace=posterior_trace,
        posterior_pickle=None,
        sample_posterior=None,
        outname=outname,
        retain_figs=False,
    )



def krusch_(
    pymc_model: Model | None = None,
    model_file: str | None = None,
    model_name: str | None = None,
    posterior_trace: InferenceData | None = None,
    posterior_pickle: str | None = None,
    sample_posterior: bool = False,
    outname: str = "krusch.png",
    retain_figs=False,
):
    
    from pykrusch.parsePymc.createGraph import create_graph
    from pykrusch.figureControl import FC


    MODEL = model_check(pymc_model, model_file, model_name)

    posterior_data: InferenceData | None = None
    plot_posterior: bool = False

    # Get pymc Model object. If directly supplied as argument to
    # MODEL, simply return MODEL. Otherwise, import the file
    # indicated in model_file and grab the model identified by
    # model_name

    plot_posterior, posterior_data = posterior_data_check(
        MODEL=MODEL,
        posterior_trace=posterior_trace,
        posterior_pickle=posterior_pickle,
        sample_posterior=sample_posterior,
    )

    # This initializes the figure numbering system
    # for the temporary figures
    FC.fig_init(FC)

    ###
    # CORE LOOP: Create graph does most of the heavy lifting
    ###
    graph: AGraph = create_graph(
        MODEL,
        plot_posterior,
        posterior_data,
    )
    graph.graph_attr["dpi"] = DPI
    graph.draw(outname, prog=ENGINE)

    if not retain_figs:
        rmtree(TEMP_IMG_DIR)

